import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.optimize import bisect
from typing import Optional, Tuple
from pysagas.flow import FlowState
from pysagas.geometry import Vector
from scipy.optimize import root_scalar
from pysagas.utilities import add_sens_data
from pysagas.cfd.solver import FlowSolver, FlowResults, SensitivityResults


class OPM(FlowSolver):
    """Oblique shock thoery and Prandtl-Meyer expansion theory flow solver.

    This implementation uses oblique shock theory for flow-facing cell
    elements, and Prandtl-Meyer expansion theory for rearward-facing elements.

    Extended Summary
    ----------------
    Data attribute 'method' refers to which method was used for a particular
    cell, according to:
        - -1 : invalid / skipped (eg. 90 degree face)
        - 0 : parallel face, do nothing
        - 1 : Prandlt-Meyer
        - 2 : normal shock
        - 3 : oblique shock
    """

    method = "Oblique/Prandtl-Meyer combination"
    PM_ANGLE_THRESHOLD = -20  # degrees

    def solve(
        self,
        freestream: Optional[FlowState] = None,
        mach: Optional[float] = None,
        aoa: Optional[float] = None,
        cog: Vector = Vector(0, 0, 0),
    ) -> FlowResults:
        already_run = super().solve(freestream=freestream, mach=mach, aoa=aoa)
        if already_run:
            # Already have a result
            if self.verbosity > 1:
                print("Attention: this flowstate has already been solved.")
            result = self.flow_result

        else:
            # Get flow
            flow = self._last_solve_freestream

            # Construct progress bar
            if self.verbosity > 0:
                print()
                desc = f"Running OPM solver at AoA = {flow.aoa:.2f} and Mach = {flow.M:.2f}"
                pbar = tqdm(
                    total=len(self.cells),
                    desc=desc,
                    position=0,
                    leave=True,
                )

            # Iterate over all cells
            net_force = Vector(0, 0, 0)
            net_moment = Vector(0, 0, 0)
            bad = 0
            total = 0
            for cell in self.cells:
                # Calculate orientation of cell to flow
                theta = np.pi / 2 - np.arccos(
                    np.dot(flow.direction.vec, cell.n.vec)
                    / (cell.n.norm * flow.direction.norm)
                )

                # Also calculate vector to COG from cell centroid
                r = cell.c - cog

                # Solve flow for this cell
                if theta < 0:
                    # Rearward facing cell
                    if theta < np.deg2rad(self.PM_ANGLE_THRESHOLD):
                        M2, p2, T2 = (flow.M, 0.0, flow.T)
                        method = -1
                        bad += 1

                    else:
                        # Use Prandtl-Meyer expansion theory
                        method = 1
                        M2, p2, T2 = self._solve_pm(
                            abs(theta), flow.M, flow.P, flow.T, flow.gamma
                        )

                elif theta > 0:
                    # Use shock theory
                    beta_max = OPM.beta_max(M=flow.M, gamma=flow.gamma)
                    theta_max = OPM.theta_from_beta(
                        M1=flow.M, beta=beta_max, gamma=flow.gamma
                    )
                    if theta > theta_max:
                        # Detached shock
                        method = 2
                        M2, p2, T2 = self._solve_normal(
                            flow.M, flow.P, flow.T, flow.gamma
                        )

                    else:
                        # Use oblique shock theory
                        method = 3
                        M2, p2, T2 = self._solve_oblique(
                            theta, flow.M, flow.P, flow.T, flow.gamma
                        )

                else:
                    # Cell is parallel to flow, assume no change
                    method = 0
                    M2, p2, T2 = (flow.M, flow.P, flow.T)

                total += 1

                # Save results for this cell
                cell.attributes["pressure"] = p2
                cell.attributes["Mach"] = M2
                cell.attributes["temperature"] = T2
                cell.attributes["method"] = method

                # Add flowstate info to Cell
                # TODO - calculate flowstate direction
                cell.flowstate = FlowState(
                    mach=M2,
                    pressure=p2,
                    temperature=T2,
                    direction=None,
                )

                # Calculate force vector
                F = cell.n * p2 * cell.A
                net_force += F
                net_moment += Vector.from_coordinates(np.cross(r.vec, F.vec))

                # Update progress bar
                if self.verbosity > 0:
                    pbar.update(1)

            if self.verbosity > 0:
                pbar.close()

                if bad / total > 0.25:
                    print(
                        f"WARNING: {100*bad/total:.2f}% of cells were not "
                        "solved due to PM threshold."
                    )

            # Construct results
            result = FlowResults(
                freestream=flow, net_force=net_force, net_moment=net_moment
            )

            # Save
            self.flow_result = result

        # Print result
        if self.verbosity > 1:
            print(result)

        return result

    @staticmethod
    def pm(M: float, gamma: float = 1.4):
        """Solves the Prandtl-Meyer function and returns the Prandtl-Meyer angle
        in radians."""
        v = ((gamma + 1) / (gamma - 1)) ** 0.5 * np.arctan(
            ((M**2 - 1) * (gamma - 1) / (gamma + 1)) ** 0.5
        ) - np.arctan((M**2 - 1) ** 0.5)
        return v

    @staticmethod
    def inv_pm(angle: float, gamma: float = 1.4):
        """Solves the inverse Prandtl-Meyer function using a bisection algorithm."""
        func = lambda M: OPM.pm(M, gamma) - angle
        pm = bisect(func, 1.0, 42.0)
        return pm

    @staticmethod
    def _solve_pm(
        theta: float, M1: float, p1: float = 1.0, T1: float = 1.0, gamma: float = 1.4
    ):
        """Solves for the Mach number, pressure and temperature after a Prandtl-Meyer
        expansion fan.

        Parameters
        ----------
        theta : float
            The deflection angle, specified in radians.

        M1 : float
            The pre-expansion Mach number.

        p1 : float, optional
            The pre-expansion pressure (Pa). The default is 1.0.

        T1 : float, optional
            The pre-expansion temperature (K). The default is 1.0.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        M2 : float
            The post-expansion Mach number.

        p2 : float
            The post-expansion pressure (Pa).

        T2 : float
            The post-expansion temperature (K).
        """
        # Solve for M2
        v_M1 = OPM.pm(M=M1, gamma=gamma)
        v_M2 = theta + v_M1
        M2 = OPM.inv_pm(v_M2)

        a = (gamma - 1) / 2
        n = 1 + a * M1**2
        d = 1 + a * M2**2

        p2 = p1 * (n / d) ** (gamma / (gamma - 1))

        T2 = T1 * (n / d)

        return M2, p2, T2

    @staticmethod
    def _solve_oblique(
        theta: float, M1: float, p1: float = 1.0, T1: float = 1.0, gamma: float = 1.4
    ):
        """Solves the flow using oblique shock theory.

        Parameters
        ----------
        theta : float
            The flow deflection angle, specified in radians.

        M1 : float
            The pre-shock Mach number.

        p1 : float, optional
            The pre-shock pressure (Pa). The default is 1.0.

        T1 : float, optional
            The pre-expansion temperature (K). The default is 1.0.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        M2 : float
            The post-shock Mach number.

        p2 : float
            The post-shock pressure (Pa).

        T2 : float
            The post-shock temperature (K).
        """
        # Calculate angle and ratios
        beta = OPM.oblique_beta(M1, theta, gamma)
        p2_p1 = OPM.oblique_p2_p1(M1, beta, gamma)
        rho2_rho1 = OPM.oblique_rho2_rho1(M1, beta, gamma)
        T2_T1 = p2_p1 / rho2_rho1

        # Calculate properties
        M2 = OPM.oblique_M2(M1, beta, theta, gamma)
        p2 = p2_p1 * p1
        T2 = T1 * T2_T1

        return M2, p2, T2

    @staticmethod
    def _solve_normal(M1: float, p1: float = 1.0, T1: float = 1.0, gamma: float = 1.4):
        """Solves the flow using normal shock theory.

        Parameters
        ----------
        M1 : float
            The pre-shock Mach number.

        p1 : float, optional
            The pre-shock pressure (Pa). The default is 1.0.

        T1 : float, optional
            The pre-expansion temperature (K). The default is 1.0.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        M2 : float
            The post-shock Mach number.

        p2 : float
            The post-shock pressure (Pa).

        T2 : float
            The post-shock temperature (K).
        """
        rho2_rho1 = (gamma + 1) * M1**2 / (2 + (gamma - 1) * M1**2)
        p2_p1 = 1 + 2 * gamma * (M1**2 - 1) / (gamma + 1)

        M2 = ((1 + M1 * (gamma - 1) / 2) / (gamma * M1**2 - (gamma - 1) / 2)) ** 0.5
        p2 = p1 * p2_p1
        T2 = T1 * p2_p1 / rho2_rho1

        return M2, p2, T2

    @staticmethod
    def oblique_p2_p1(M1: float, beta: float, gamma: float = 1.4):
        """Returns the pressure ratio p2/p1 across an oblique shock.

        Parameters
        -----------
        M1 : float
            The pre-shock Mach number.

        beta : float
            The shock angle specified in radians.

        Returns
        --------
        p2/p1 : float
            The pressure ratio across the shock.

        References
        ----------
        Peter Jacobs
        """
        M1n = M1 * abs(np.sin(beta))
        p2p1 = 1.0 + 2.0 * gamma / (gamma + 1.0) * (M1n**2 - 1.0)
        return p2p1

    @staticmethod
    def oblique_beta(
        M1: float, theta: float, gamma: float = 1.4, tolerance: float = 1.0e-6
    ):
        """Calculates the oblique shock angle using a recursive algorithm.

        Parameters
        ----------
        M1 : float
            The pre-shock Mach number.

        theta : float
            The flow deflection angle, specified in radians.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        References
        ----------
        Peter Jacobs
        """
        func = lambda beta: OPM.theta_from_beta(M1, beta, gamma) - theta

        # Initialise
        sign_beta = np.sign(theta)
        theta = abs(theta)
        b1 = np.arcsin(1.0 / M1)
        b2 = b1 * 1.05

        # Check f1
        f1 = func(b1)
        if abs(f1) < tolerance:
            return sign_beta * b1

        # Check f2
        f2 = func(b2)
        if abs(f2) < tolerance:
            return sign_beta * b2

        # Finally, solve with secant method
        # beta = sign_beta * secant(func, b1, b2, tol=tolerance)
        root_result = root_scalar(f=func, x0=b1, x1=b2, xtol=tolerance, method="secant")
        if root_result.converged:
            beta = sign_beta * root_result.root
        else:
            raise Exception(
                f"Cannot converge beta (theta={np.rad2deg(theta):.2f} degrees)."
            )

        return beta

    @staticmethod
    def theta_from_beta(M1: float, beta: float, gamma: float = 1.4):
        """Calculates the flow deflection angle from the oblique shock angle.

        Parameters
        ----------
        M1 : float
            The pre-expansion Mach number.

        beta : float
            The shock angle specified in radians.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        theta : float
            The deflection angle, specified in radians.

        References
        ----------
        Peter Jacobs
        """
        M1n = M1 * abs(np.sin(beta))
        t1 = 2.0 / np.tan(beta) * (M1n**2 - 1)
        t2 = M1**2 * (gamma + np.cos(2 * beta)) + 2
        theta = np.arctan(t1 / t2)
        return theta

    @staticmethod
    def oblique_M2(M1: float, beta: float, theta: float, gamma: float = 1.4):
        """Calculates the Mach number following an oblique shock.

        Parameters
        ----------
        M1 : float
            The pre-expansion Mach number.

        beta : float
            The shock angle specified in radians.

        theta : float
            The deflection angle, specified in radians.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        M2 : float
            The post-shock Mach number.

        References
        ----------
        Peter Jacobs
        """
        M1n = M1 * abs(np.sin(beta))
        a = 1 + (gamma - 1) * 0.5 * M1n**2
        b = gamma * M1n**2 - (gamma - 1) * 0.5
        M2 = (a / b / (np.sin(beta - theta)) ** 2) ** 0.5
        return M2

    def oblique_T2_T1(M1: float, beta: float, gamma: float = 1.4):
        """Returns the temperature ratio T2/T1 across an oblique shock.

        Parameters
        -----------
        M1 : float
            The pre-shock Mach number.

        beta : float
            The shock angle specified in radians.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        T2/T1 : float
            The temperature ratio across the shock.

        References
        ----------
        Peter Jacobs
        """
        T2_T1 = OPM.oblique_p2_p1(M1, beta, gamma) / OPM.oblique_rho2_rho1(
            M1, beta, gamma
        )
        return T2_T1

    @staticmethod
    def oblique_rho2_rho1(M1: float, beta: float, gamma: float = 1.4):
        """Returns the density ratio rho2/rho1 across an oblique shock.

        Parameters
        -----------
        M1 : float
            The pre-shock Mach number.

        beta : float
            The shock angle specified in radians.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        Returns
        --------
        T2/T1 : float
            The temperature ratio across the shock.

        References
        ----------
        Peter Jacobs
        """
        M1n = M1 * abs(np.sin(beta))
        rho2_rho1 = (gamma + 1) * M1n**2 / 2 + (gamma - 1) * M1n**2
        return rho2_rho1

    @staticmethod
    def beta_max(M: float, gamma: float = 1.4):
        """Returns the maximum shock angle for a given
        Mach number.
        """
        beta_max = np.arcsin(
            np.sqrt(
                (1 / (gamma * M**2))
                * (
                    (gamma + 1) * M**2 / 4
                    - 1
                    + np.sqrt(
                        (gamma + 1)
                        * ((gamma + 1) * M**4 / 16 + (gamma - 1) * M**2 / 2 + 1)
                    )
                )
            )
        )
        return beta_max

    def solve_sens(
        self,
        sensitivity_filepath: Optional[str] = None,
        parameters: Optional[list[str]] = None,
        cells_have_sens_data: Optional[bool] = False,
        freestream: Optional[FlowState] = None,
        mach: Optional[float] = None,
        aoa: Optional[float] = None,
        cog: Vector = Vector(0, 0, 0),
        perturbation: float = 1e-3,
    ) -> SensitivityResults:
        # TODO - add to cell attributes for visualisation
        # TODO - return Mach and Temp sensitivities

        already_run = super().solve_sens(freestream=freestream, mach=mach, aoa=aoa)
        if already_run:
            # Already have a result
            if self.verbosity > 1:
                print("Attention: this flowstate sensitivity has already been solved.")
            result = self.flow_sensitivity

        else:
            # Get flow
            flow = self._last_sens_freestream

            # Check that nominal flow condition has been solved already
            if self._last_solve_freestream:
                # Solver has run before
                if self._last_solve_freestream != flow:
                    # Run flow solver first at the nominal flow conditions
                    if self.verbosity > 0:
                        print("Running flow solver to prepare sensitivities.")
                    self.solve(flow)
            else:
                # Solver has not run yet at all, run it now
                self.solve(flow)

            if cells_have_sens_data and not parameters:
                # Need to extract parameters
                _, parameters = self._load_sens_data(sensitivity_filepath)
            elif not cells_have_sens_data:
                # Load sensitivity data and parameters
                sensdata, parameters = self._load_sens_data(sensitivity_filepath)

                # Add sensitivity data to cells
                add_sens_data(cells=self.cells, data=sensdata, verbosity=self.verbosity)

            # Construct progress bar
            if self.verbosity > 0:
                print()
                desc = "Solving cell flow sensitivities."
                pbar = tqdm(
                    total=len(self.cells),
                    desc=desc,
                    position=0,
                    leave=True,
                )

            # Iterate over cells
            dFdp = 0
            dMdp = 0
            for cell in self.cells:
                # Initialisation
                all_directions = [Vector(1, 0, 0), Vector(0, 1, 0), Vector(0, 0, 1)]
                sensitivities = np.empty(shape=(cell.dndp.shape[1], 3))
                moment_sensitivities = np.empty(shape=(cell.dndp.shape[1], 3))

                # Calculate moment dependencies
                r = cell.c - cog
                F = cell.flowstate.P * cell.A * cell.n.vec

                # Calculate orientation of cell to flow
                theta = np.pi / 2 - np.arccos(
                    np.dot(flow.direction.vec, cell.n.vec)
                    / (cell.n.norm * flow.direction.norm)
                )

                # Solve flow for this cell
                if theta < 0:
                    # Rearward facing cell
                    if theta < np.deg2rad(self.PM_ANGLE_THRESHOLD):
                        _, dP_dtheta, _ = (0.0, 0.0, 0.0)

                    else:
                        # Use Prandtl-Meyer expansion theory
                        _, dP_dtheta, _ = self.dp_dtheta_pm(
                            abs(theta),
                            flow.M,
                            cell.flowstate,
                            flow.P,
                            flow.T,
                            flow.gamma,
                            perturbation,
                        )

                elif theta > 0:
                    # Use shock theory
                    beta_max = OPM.beta_max(M=flow.M, gamma=flow.gamma)
                    theta_max = OPM.theta_from_beta(
                        M1=flow.M, beta=beta_max, gamma=flow.gamma
                    )
                    if theta > theta_max:
                        # Detached shock
                        _, dP_dtheta, _ = (0.0, 0.0, 0.0)

                    else:
                        # Use oblique shock theory
                        _, dP_dtheta, _ = self.dp_dtheta_obl(
                            theta,
                            flow.M,
                            cell.flowstate,
                            flow.P,
                            flow.T,
                            flow.gamma,
                            perturbation,
                        )

                else:
                    # Cell is parallel to flow, assume no change
                    _, dP_dtheta, _ = (0.0, 0.0, 0.0)

                # For each parameter
                for p_i in range(cell.dndp.shape[1]):
                    # Calculate gradient of pressure with respect to parameter
                    dtheta_dp = (
                        -np.dot(cell.dndp[:, p_i], flow.direction.vec)
                        / (1 - np.dot(cell.n.unit.vec, flow.direction.unit.vec) ** 2)
                        ** 0.5
                    )

                    if np.isnan(dtheta_dp):
                        # Normal vector parallel to flow
                        dtheta_dp = 0

                    dPdp = dP_dtheta * dtheta_dp

                    # Calculate pressure sensitivity for each direction
                    for i, direction in enumerate(all_directions):
                        dF = (
                            dPdp * cell.A * np.dot(cell.n.vec, direction.vec)
                            + cell.flowstate.P
                            * cell.dAdp[p_i]
                            * np.dot(cell.n.vec, direction.vec)
                            + cell.flowstate.P
                            * cell.A
                            * np.dot(-cell.dndp[:, p_i], direction.vec)
                        )
                        sensitivities[p_i, i] = dF

                    # Now evaluate moment sensitivities
                    moment_sensitivities[p_i, :] = np.cross(
                        r.vec, sensitivities[p_i, :]
                    ) + np.cross(cell.dcdp[:, p_i], F)

                # Append to cell
                cell.sensitivities = sensitivities
                cell.moment_sensitivities = moment_sensitivities

                # Update
                dFdp += sensitivities
                dMdp += moment_sensitivities

                # Update progress bar
                if self.verbosity > 0:
                    pbar.update(1)

            if self.verbosity > 0:
                pbar.close()
                print("Done.")

            # Construct dataframes to return
            df_f = pd.DataFrame(
                dFdp, columns=["dFx/dp", "dFy/dp", "dFz/dp"], index=parameters
            )
            df_m = pd.DataFrame(
                dMdp, columns=["dMx/dp", "dMy/dp", "dMz/dp"], index=parameters
            )

            # Construct results
            result = SensitivityResults(
                f_sens=df_f,
                m_sens=df_m,
                freestream=flow,
            )

            # Save
            self.flow_sensitivity = result

        if self.verbosity > 0:
            print(result)

        return result

    @staticmethod
    def dp_dtheta_obl(
        theta: float,
        M1: float,
        nominal_flowstate: FlowState,
        p1: float = 1.0,
        T1: float = 1.0,
        gamma: float = 1.4,
        perturbation: float = 1e-3,
    ):
        """Solves for the sensitivities at the given point using oblique shock
        theory.

        Parameters
        ----------
        theta : float
            The deflection angle, specified in radians.

        M1 : float
            The pre-expansion Mach number.

        nominal_flowstate : FlowState
            The nominal flowstate at this point.

        p1 : float, optional
            The pre-expansion pressure (Pa). The default is 1.0.

        T1 : float, optional
            The pre-expansion temperature (K). The default is 1.0.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        perturbation : float, optional
            The perturbation amount. Perturbations are calculated
            according to a multiple of (1 +/- perturbation). The
            default is 1e-3.

        Returns
        --------
        dM_dtheta : float
            The sensitivity of the post-expansion Mach number with respect
            to the deflection angle theta.

        dP_dtheta : float
            The sensitivity of the post-expansion pressure (Pa) with respect
            to the deflection angle theta.

        dT_dtheta : float
            The sensitivity of the post-expansion temperature (K) with respect
            to the deflection angle theta.
        """
        # Create function handle
        func = lambda theta: OPM._solve_oblique(theta, M1, p1, T1, gamma)

        # Calculate derivitive by finite differencing
        g = OPM._findiff(func, theta, nominal_flowstate, perturbation)

        return g

    @staticmethod
    def dp_dtheta_pm(
        theta: float,
        M1: float,
        nominal_flowstate: FlowState,
        p1: float = 1.0,
        T1: float = 1.0,
        gamma: float = 1.4,
        perturbation: float = 1e-3,
    ):
        """Solves for the sensitivities at the given point using Prandtl-Meyer
        theory.

        Parameters
        ----------
        theta : float
            The deflection angle, specified in radians.

        M1 : float
            The pre-expansion Mach number.

        nominal_flowstate : FlowState
            The nominal flowstate at this point.

        p1 : float, optional
            The pre-expansion pressure (Pa). The default is 1.0.

        T1 : float, optional
            The pre-expansion temperature (K). The default is 1.0.

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.

        perturbation : float, optional
            The perturbation amount. Perturbations are calculated
            according to a multiple of (1 +/- perturbation). The
            default is 1e-3.

        Returns
        --------
        dM_dtheta : float
            The sensitivity of the post-expansion Mach number with respect
            to the deflection angle theta.

        dP_dtheta : float
            The sensitivity of the post-expansion pressure (Pa) with respect
            to the deflection angle theta.

        dT_dtheta : float
            The sensitivity of the post-expansion temperature (K) with respect
            to the deflection angle theta.
        """
        # Create function handle
        func = lambda theta: OPM._solve_pm(theta, M1, p1, T1, gamma)

        # Calculate derivitive by finite differencing
        g = OPM._findiff(func, theta, nominal_flowstate, perturbation)

        return g

    @staticmethod
    def _findiff(
        func: callable,
        point: any,
        nominal_flowstate: FlowState,
        perturbation: float = 1e-3,
    ):
        # Generate flow value at points +/- perturbed from nominal point
        xvals = [point * (1 - perturbation), point, point * (1 + perturbation)]
        vals = [
            func(p) for p in [point * (1 - perturbation), point * (1 + perturbation)]
        ]
        Mvals = [v[0] for v in vals]
        pvals = [v[1] for v in vals]
        Tvals = [v[2] for v in vals]

        # Inject nominal result
        Mvals = [Mvals[0], nominal_flowstate.M, Mvals[1]]
        pvals = [pvals[0], nominal_flowstate.P, pvals[1]]
        Tvals = [Tvals[0], nominal_flowstate.T, Tvals[1]]

        dM = np.gradient(Mvals, xvals)
        dp = np.gradient(pvals, xvals)
        dT = np.gradient(Tvals, xvals)

        return dM[1], dp[1], dT[1]

    def save(self, name: str):
        # Initialise attributes dictionary
        attributes = {
            "pressure": [],
            "Mach": [],
            "temperature": [],
            "method": [],
        }

        # Call super method
        super().save(name, attributes)

    @staticmethod
    def _load_sens_data(sensitivity_filepath: str) -> Tuple[pd.DataFrame, list[set]]:
        """Extracts design parameters from the sensitivity data."""
        sensdata = pd.read_csv(sensitivity_filepath)

        parameters = set()
        for e in sensdata.columns:
            e: str
            if e.startswith("dxd") or e.startswith("dyd") or e.startswith("dzd"):
                # Sensitivity coluns
                parameters.add(e[3:])

        return sensdata, list(parameters)



import numpy as np
import pandas as pd
from pysagas.flow import FlowState
from abc import ABC, abstractmethod
from pysagas.geometry import Vector, Cell
from pysagas.utilities import add_sens_data
from pysagas.cfd.solver import SensitivityResults
from typing import List, Callable, Tuple, Optional, Union, Literal
from pysagas.sensitivity.models import (
    piston_sensitivity,
    van_dyke_sensitivity,
    isentropic_sensitivity,
)


class AbstractSensitivityCalculator(ABC):
    """Abstract sensitivity calculator class defining the interface."""

    solver = None

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        pass

    def __repr__(self) -> str:
        return f"PySAGAS Sensitivity Calculator for {self.solver}"

    def __str__(self) -> str:
        return f"PySAGAS Sensitivity Calculator for {self.solver}"

    @property
    @abstractmethod
    def solver(self):
        # This is a placeholder for a class variable defining the Sensitivity Calculator's solver
        pass

    @abstractmethod
    def _transcribe_cells(self, parameters: List):
        """Transcribes the unified Cells from the solver data.

        Returns
        --------
        cells : List[Cell]
            A list of the Cells from the flow solution.
        """
        # This method should be overloaded with the solver-specific method
        pass

    @abstractmethod
    def _extract_parameters(self) -> List[str]:
        """Extract the geometry parameters.

        Returns
        -------
        parameters : List[str]
            A list of the parameter names.
        """
        # This method should be overloaded with the solver-specific method
        pass

    @abstractmethod
    def to_csv(self):
        """Dumps the sensitivity data to CSV file."""
        pass


class SensitivityCalculator(AbstractSensitivityCalculator):
    """SensitivityCalculator base class."""

    def __init__(self, **kwargs) -> None:
        self.cells: list[Cell] = None
        self.verbosity = 1

    def calculate(
        self,
        sensitivity_model: Optional[
            Union[Callable, Literal["piston", "van_dyke", "isentropic"]]
        ] = "van_dyke",
        cog: Optional[Vector] = Vector(0, 0, 0),
        flowstate: Optional[FlowState] = None,
        **kwargs,
    ) -> SensitivityResults:
        """Calculate the force sensitivities of the surface to the
        parameters.

        Parameters
        ----------
        sensitivity_model : Callable | Literal["piston", "van_dyke", "isentropic"], optional
            The model used to calculate the pressure/parameter sensitivities.
            The default is Van Dyke's second-order theory model.

        cog : Vector, optional
            The centre of gravity. The default is Vector(0, 0, 0).

        flowstate : FlowState, optional
            The flowstate associated with this result. The default is None.

        Returns
        -------
        SensitivityResults : the sensitivity results object.
        """
        if self.verbosity > 0:
            print("\nCalculating aerodynamic sensitivities.")

        parameters = self._extract_parameters()

        if self.cells is None:
            self.cells = self._transcribe_cells(parameters=parameters)

        params_sens_cols = []
        for p in parameters:
            for d in ["x", "y", "z"]:
                params_sens_cols.append(f"d{d}d_{p}")

        # Calculate force sensitivity
        F_sense, M_sense = self.net_sensitivity(
            cells=self.cells, sensitivity_model=sensitivity_model, cog=cog, **kwargs
        )

        # Construct dataframes to return
        df_f = pd.DataFrame(
            F_sense, columns=["dFx/dp", "dFy/dp", "dFz/dp"], index=parameters
        )
        df_m = pd.DataFrame(
            M_sense, columns=["dMx/dp", "dMy/dp", "dMz/dp"], index=parameters
        )

        if self.verbosity > 0:
            print("Done.")

        # Construct results
        result = SensitivityResults(
            f_sens=df_f,
            m_sens=df_m,
            freestream=flowstate,
        )

        return result

    def to_csv(self):
        """Dumps the sensitivity data to CSV file."""
        if self.cells is not None:
            pass

        else:
            raise Exception("No cells have been transcribed.")

    def _extract_parameters(self):
        parameters = []
        for e in self.sensdata.columns:
            e: str
            if e.startswith("dxd") or e.startswith("dyd") or e.startswith("dzd"):
                # Sensitivity coluns
                if e[3:] not in parameters:
                    parameters.append(e[3:])
        return parameters

    @staticmethod
    def net_sensitivity(
        cells: List[Cell],
        sensitivity_model: Optional[
            Union[Callable, Literal["piston", "van_dyke", "isentropic"]]
        ] = "van_dyke",
        cog: Vector = Vector(0, 0, 0),
        **kwargs,
    ) -> Tuple[np.array, np.array]:
        """Calcualtes the net force and moment sensitivities for a list of Cells.

        Parameters
        ----------
        cells : list[Cell]
            The cells to be analysed.

        sensitivity_model : Callable | Literal["piston", "van_dyke", "isentropic"], optional
            The model used to calculate the pressure/parameter sensitivities.
            The default is Van Dyke's second-order theory model.

        cog : Vector, optional
            The reference centre of gravity, used in calculating the moment
            sensitivities. The defualt is Vector(0,0,0).

        Returns
        --------
        dFdp : np.array
            The force sensitivity matrix with respect to the parameters.

        dMdp : np.array
            The moment sensitivity matrix with respect to the parameters.
        """
        # Get sensitivity handle
        if isinstance(sensitivity_model, Callable):
            # Use function provided
            sensitivity_function = sensitivity_model
        elif isinstance(sensitivity_model, str):
            # Get function from method specified
            sensitivity_function = globals().get(f"{sensitivity_model}_sensitivity")
            if sensitivity_function is None:
                raise Exception("Invalid sensitivity method specified.")

        dFdp = 0
        dMdp = 0
        for cell in cells:
            # Calculate force sensitivity
            dFdp_c, dMdp_c = SensitivityCalculator.cell_sensitivity(
                cell=cell, sensitivity_function=sensitivity_function, cog=cog, **kwargs
            )

            dFdp += dFdp_c
            dMdp += dMdp_c

        return dFdp, dMdp

    @staticmethod
    def cell_sensitivity(
        cell: Cell,
        sensitivity_function: Callable,
        cog: Vector = Vector(0, 0, 0),
        **kwargs,
    ) -> Tuple[np.array, np.array]:
        """Calculates force and moment sensitivities for a single cell.

        Parameters
        ----------
        cell : Cell
            The cell.

        sensitivity_function : Callable
            The function to use when calculating the pressure sensitivities.

        cog : Vector, optional
            The reference centre of gravity, used in calculating the moment
            sensitivities. The defualt is Vector(0,0,0).

        Returns
        --------
        sensitivities : np.array
            An array of shape n x 3, for a 3-dimensional cell with
            n parameters.

        See Also
        --------
        all_dfdp : a wrapper to calculate force sensitivities for many cells
        """
        # Initialisation
        all_directions = [Vector(1, 0, 0), Vector(0, 1, 0), Vector(0, 0, 1)]
        sensitivities = np.zeros(shape=(cell.dndp.shape[1], 3))
        moment_sensitivities = np.zeros(shape=(cell.dndp.shape[1], 3))

        # Calculate moment dependencies
        r = cell.c - cog
        F = cell.flowstate.P * cell.A * cell.n.vec

        # For each parameter
        for p_i in range(cell.dndp.shape[1]):
            # Calculate pressure sensitivity
            dPdp = sensitivity_function(cell=cell, p_i=p_i, **kwargs)

            # Evaluate for sensitivities for each direction
            for i, direction in enumerate(all_directions):
                dF = (
                    dPdp * cell.A * np.dot(cell.n.vec, direction.vec)
                    + cell.flowstate.P
                    * cell.dAdp[p_i]
                    * np.dot(cell.n.vec, direction.vec)
                    + cell.flowstate.P
                    * cell.A
                    * np.dot(-cell.dndp[:, p_i], direction.vec)
                )
                sensitivities[p_i, i] = dF

            # Now evaluate moment sensitivities
            moment_sensitivities[p_i, :] = np.cross(
                r.vec, sensitivities[p_i, :]
            ) + np.cross(cell.dcdp[:, p_i], F)

        # Append to cell
        cell.sensitivities = sensitivities
        cell.moment_sensitivities = moment_sensitivities

        return sensitivities, moment_sensitivities


class GenericSensitivityCalculator(SensitivityCalculator):
    solver = "Generic Flow Solver"

    def __init__(
        self,
        cells: List[Cell],
        sensitivity_filepath: Optional[str] = None,
        cells_have_sens_data: Optional[bool] = False,
        verbosity: Optional[int] = 1,
        **kwargs,
    ) -> None:
        """Instantiate a generic PySAGAS sensitivity calculator.

        Parameters
        ----------
        cells : list[Cell]
            A list of transcribed Cell objects, containing the nominal flow solution.

        sensitivity_filepath : str
            The filepath to the geometry sensitivities. This must be provided, if the
            cells do not have geometric sensitivity information attached. This must
            also be provided anyway, to determine the geometric design parameters.

        cells_have_sens_data : bool, optional
            The cells already have geometric sensitivity data matched to them. When this
            is True, the sensitivity_filepath argument (if provided) is ignored. The
            default is False.

        verbosity : int, optional
            The verbosity of the code. The defualt is 1.
        """
        super().__init__(**kwargs)

        if cells_have_sens_data:
            self.cells = cells
        else:
            self._pre_transcribed_cells = cells

        self.sensdata = pd.read_csv(sensitivity_filepath)
        self.verbosity = verbosity

    def _transcribe_cells(self, **kwargs) -> List[Cell]:
        """This is a dummy method to satisfy the abstract base class. Transcribed cells
        are provided upon instantiation of the sensitivity calculator.
        """
        # Add sensitivity data to _pre_transcribed_cells
        add_sens_data(
            cells=self._pre_transcribed_cells,
            data=self.sensdata,
            verbosity=self.verbosity,
        )

        return self._pre_transcribed_cells



import os as os
import numpy as np
import pandas as pd
from pysagas.flow import FlowState
from pysagas.geometry import Vector
from pysagas.sensitivity import Cart3DSensitivityCalculator
from pysagas.sensitivity.models import van_dyke_sensitivity


def run_main(data_path):
    """Analyse the Cart3D solution."""
    L_ref = 1  # m
    A_ref = 1  # m2
    cog = Vector(2.25, 0, 0)

    # Create freestream flowstate
    freestream = FlowState(mach=6, pressure=1.975e3, temperature=223)

    # Filepaths
    sensitivity_filepath = os.path.join(data_path, "combined.csv")

    # Data
    pointdata = pd.read_csv(os.path.join(data_path, "points.csv"))
    celldata = pd.read_csv(os.path.join(data_path, "cells.csv"))

    # Create PySAGAS sensitivity calculator
    wrapper = Cart3DSensitivityCalculator(
        freestream=freestream,
        sensitivity_filepath=sensitivity_filepath,
        pointdata=pointdata,
        celldata=celldata,
    )

    vd_result = wrapper.calculate(
        cog=cog,
        sensitivity_model=van_dyke_sensitivity,
        freestream=freestream,
        dp=[0.1 * 0.05],
    )

    coef_sensv = vd_result.f_sens / (freestream.q * A_ref)
    _ = vd_result.m_sens / (freestream.q * A_ref * L_ref)

    # Calculate sensitivities
    piston_result = wrapper.calculate(cog=cog, sensitivity_model="piston")

    # Non-dimensionalise
    coef_sens = piston_result.f_sens / (freestream.q * A_ref)
    _ = piston_result.m_sens / (freestream.q * A_ref * L_ref)

    print("\nCart 3D Finite Difference Result:")
    c3d_sens = np.array([[0.147517, 0.126153, 0]])
    print(c3d_sens)

    # Print results
    print("\nPySAGAS Result:")
    print("Piston theory:")
    print(coef_sens)
    print("Van Dykes theory:")
    print(coef_sensv)

    print("\nError (%):")
    errors = np.nan_to_num(100 * (coef_sens - c3d_sens) / c3d_sens, posinf=0, neginf=0)
    print(errors)

    return errors


def cart3d_fd(data_path):
    """Calculate differences from Cart3D result using finite
    differencing."""
    # Load data
    cols = {
        "thickness": "thickness",
        "C_D-entire": "CD",
        "C_L-entire": "CL",
        "C_m-entire": "Cm",
    }
    data = pd.read_csv(os.path.join(data_path, "sims.csv"), usecols=cols.keys())
    data.rename(columns=cols, inplace=True)

    # Transform into CX, CY and CZ
    aoa = np.deg2rad(3)
    data["CX"] = -data["CL"] * np.sin(aoa) + data["CD"] * np.cos(aoa)
    data["CY"] = data["CL"] * np.cos(aoa) + data["CD"] * np.sin(aoa)

    # Calculate sensitivities
    pc1 = (data.iloc[4] - data.iloc[3]) / (
        data.iloc[4]["thickness"] - data.iloc[3]["thickness"]
    )

    return pc1


def test_cart3d_wedge():
    """Run the test."""
    file_dir = os.path.abspath(os.path.dirname(__file__))
    data_path = os.path.join(file_dir, "data")
    errors = run_main(data_path)

    assert np.max(np.abs(errors)) < 5.6, "Sensitivities inaccurately calculated"


if __name__ == "__main__":
    # cart3d_fd("tests/data")
    np.seterr(all="ignore")
    data_path = "tests/data"
    run_main(data_path)



import numpy as np
from pysagas.cfd import OPM


def test_pm():
    """Test Prandtl-Meyer solver."""
    M1 = 1.5
    p1 = 101e3
    T1 = 288
    theta = np.deg2rad(15)
    M2, p2, T2 = OPM._solve_pm(theta, M1, p1, T1)

    assert round(M2) == 2.0, f"incorrect M2 (M2={M2})"
    assert round(p2) == 45998, f"incorrect p2 (p2={p2})"
    assert round(T2) == 230, f"incorrect T2 (T2={T2})"



from __future__ import annotations
import numpy as np
from typing import Union


class Vector:
    """An N-dimensional vector."""

    PRECISION = 3

    def __init__(self, x: float = None, y: float = None, z: float = None):
        """Define a new vector.

        Parameters
        ----------
        x : float, optional
            The x-component of the vector.

        y : float, optional
            The y-component of the vector.

        z : float, optional
            The z-component of the vector.
        """
        self._x = x
        self._y = y
        self._z = z

        self._non_none = [str(i) for i in [x, y, z] if i is not None]
        self._round_non_none = [
            str(round(i, self.PRECISION)) for i in [x, y, z] if i is not None
        ]
        self._dimensions = len(self._non_none)

    def __str__(self) -> str:
        s = f"{self._dimensions}-dimensional vector: ({', '.join(self._round_non_none)})"
        return s

    def __repr__(self) -> str:
        return f"Vector({', '.join(self._round_non_none)})"

    def __neg__(self):
        """Returns the vector pointing in the opposite direction."""
        return Vector(x=-1 * self.x, y=-1 * self.y, z=-1 * self.z)

    def __add__(self, other):
        """Element-wise vector addition.

        Parameters
        ----------
        other : Vector
            Another Vector object to be added. This Vector must be of the same
            dimension as the one it is being added to.
        """
        if not isinstance(other, Vector):
            raise Exception(f"Cannot add a {type(other)} to a vector.")
        return Vector(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)

    def __sub__(self, other):
        """Element-wise vector subtraction.

        Parameters
        ----------
        other : Vector
            Another Vector object to be added. This Vector must be of the same
            dimension as the one it is being added to.
        """
        if not isinstance(other, Vector):
            raise Exception(f"Cannot add a {type(other)} to a vector.")
        return Vector(x=self.x - other.x, y=self.y - other.y, z=self.z - other.z)

    def __truediv__(self, denominator: Union[float, int]):
        """Element-wise vector division.

        Parameters
        ----------
        denominator : float | int
            The denominator to use in the division.
        """
        return Vector(
            x=self.x / denominator, y=self.y / denominator, z=self.z / denominator
        )

    def __mul__(self, multiple: Union[float, int]):
        """Element-wise vector multiplication.

        Parameters
        ----------
        multiple : float | int
            The multiple to use in the multiplication.
        """
        return Vector(x=self.x * multiple, y=self.y * multiple, z=self.z * multiple)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Vector):
            raise Exception(f"Cannot compare a {type(other)} to a vector.")
        return (self.x == other.x) & (self.y == other.y) & (self.z == other.z)

    @property
    def x(self) -> float:
        return self._x

    @property
    def y(self) -> float:
        return self._y

    @property
    def z(self) -> float:
        return self._z

    @property
    def vec(self) -> np.array:
        """The vector represented as a Numpy array."""
        return np.array([float(i) for i in self._non_none])

    @property
    def unit(self) -> Vector:
        """The unit vector associated with the Vector."""
        return self / self.norm

    @property
    def norm(self) -> Vector:
        """The norm associated with the Vector."""
        return np.linalg.norm(self.vec)

    @classmethod
    def from_coordinates(cls, coordinates: np.array) -> Vector:
        """Constructs a Vector object from an array of coordinates.

        Parameters
        ----------
        coordinates : np.array
            The coordinates of the vector.

        Returns
        -------
        Vector

        Examples
        --------
        >>> Vector.from_coordinates([1,2,3])
        Vector(1, 2, 3)
        """
        return cls(*coordinates)



import codecs
import os.path
import setuptools


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


# Define extra dependencies
dev_dep = ["pytest >= 7.1.1", "black >= 22.10.0", "commitizen >= 2.35.0", "pre-commit"]
docs_dep = [
    "sphinx == 5.3.0",
    "sphinx-rtd-theme == 1.1.1",
    "sphinx-copybutton >= 0.5.0",
    "sphinx-inline-tabs >= 2022.1.2b11",
    "myst-parser >= 0.18.1",
    "sphinx-autobuild >= 2021.3.14",
]
all_dep = dev_dep + docs_dep

# Load README
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hypysagas",
    version=get_version("pysagas/__init__.py"),
    author="Kieran Mackle",
    author_email="kemackle98@gmail.com",
    description="Python Sensitivity Approximations for Geometric-Aerodynamic Surfaces",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        "Bug Tracker": "https://github.com/kieran-mackle/pysagas/issues",
        "Source Code": "https://github.com/kieran-mackle/pysagas",
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.7",
    install_requires=[
        "numpy >= 1.23.5",
        "pandas >= 1.5.2",
        "tqdm >= 4.64.1",
        "art >= 5.8",
        "meshio >= 5.3.4",
        "matplotlib",
        "scipy >= 1.10.1",
    ],
    extras_require={
        "dev": dev_dep,
        "all": all_dep,
    },
    setup_requires=[
        "setuptools_git",
        "setuptools_scm",
    ],
)
